/*
 * Copyright (c) 2003-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/ats_sscfu/user.c,v 1.2 2004/01/02 15:15:54 hbb Exp $
 */
#include <sys/types.h>
#include <sys/queue.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <limits.h>
#include <err.h>
#include <isc/eventlib.h>

#include <netnatm/unimsg.h>
#include <netnatm/saal/sscopdef.h>
#include <netnatm/saal/sscop.h>
#include <netnatm/saal/sscfudef.h>
#include <netnatm/saal/sscfu.h>

#include "ats.h"
#include "ats_sscfu.h"

struct timer {
	evTimerID id;
	struct sscop *sscop;
	void (*func)(void *);
};

static struct glue {
	struct sscop *sscop;
	struct sscfu *sscf;
	evContext *ctx;
} stack;

static u_int debug_flags;

/*
 * Set debugging
 */
static void
user_debug(u_int flags)
{
	debug_flags = flags;
}

/*
 * Reset the stack
 */
static void
user_reset(void)
{

	sscop_reset(stack.sscop);
	sscfu_reset(stack.sscf);
}

/*
 * Start the stack
 */
static void
user_start(void)
{

	sscop_setdebug(stack.sscop, debug_flags);
}

/*
 * timer helper
 */
static void
tfunc(evContext ctx __unused, void *uap, struct timespec due __unused,
    struct timespec inter __unused)
{
	struct timer *t = uap;

	t->func(t->sscop);
	free(t);
}


/*
 * Start an uni timer
 */
static void *
sscop_start_timer(struct sscop *sscop, void *arg, u_int msec,
    void (*func)(void *))
{
	struct timer *t;
	struct timespec due, now, ms;
	struct glue *glue = arg;

	if ((t = malloc(sizeof(*t))) == NULL)
		err(1, NULL);
	t->sscop = sscop;
	t->func = func;
	evInitID(&t->id);

	now = evNowTime();
	ms = evConsTime((time_t)(msec / 1000), (long)(msec % 1000) * 1000000);
	due = evAddTime(now, ms);

	if (evSetTimer(*glue->ctx, tfunc, t, due, evConsTime(0, 0), &t->id))
		err(1, "cannot start timer");

	return (t);
}

/*
 * Stop an uni timer
 */
static void
sscop_stop_timer(struct sscop *sscop __unused, void *arg, void *tp)
{
	struct glue *glue = arg;
	struct timer *t = tp;

	evClearTimer(*glue->ctx, t->id);
	free(t);
}

/*
 * output to the lower layer
 */
static void
sscop_send_lower(struct sscop *sscop __unused, void *arg __unused,
    struct uni_msg *msg)
{
	struct winput *ip;

	if ((ip = malloc(sizeof(*ip))) == NULL)
		err(1, NULL);

	ip->msg = msg;
	TAILQ_INSERT_TAIL(&lower_input_list, ip, link);
}

/*
 * output to the upper layer
 */
static void
sscfu_send_upper(struct sscfu *sscfu __unused, void *arg __unused,
    enum saal_sig sig, struct uni_msg *msg)
{
	struct winput *ip;

	if ((ip = malloc(sizeof(*ip))) == NULL)
		err(1, NULL);

	ip->msg = msg;
	ip->sig = sig;
	TAILQ_INSERT_TAIL(&upper_input_list, ip, link);
}

/*
 * output to the manage layer
 */
static void
sscop_send_manage(struct sscop *sscop __unused, void *arg __unused,
    enum sscop_maasig sig, struct uni_msg *msg, u_int sarg1, u_int sarg2)
{
	struct winput *ip;

	if ((ip = malloc(sizeof(*ip))) == NULL)
		err(1, NULL);

	ip->msg = msg;
	ip->sig = sig;
	ip->arg1 = sarg1;
	ip->arg2 = sarg2;
	TAILQ_INSERT_TAIL(&manage_input_list, ip, link);
}

/*
 * SSCOP -> SSCF
 */
static void
sscop_send_upper(struct sscop *sscop __unused, void *arg, enum sscop_aasig sig,
    struct uni_msg *msg, u_int sigarg)
{

	sscfu_input(((struct glue *)arg)->sscf, sig, msg, sigarg);
}

/*
 * SSCF -> SSCOP
 */
static void
sscfu_send_lower(struct sscfu *sscfu __unused, void *arg,
    enum sscop_aasig sig, struct uni_msg *msg, u_int sarg)
{


	if (sscop_aasig(((struct glue *)arg)->sscop, sig, msg, sarg))
		errx(1, "sscop_aasig returns error");
}

/*
 * SSCOP window update
 */
static void
sscfu_window(struct sscfu *sscfu __unused, void *arg, u_int n)
{

	sscop_window(((struct glue *)arg)->sscop, n);
}

/*
 * verbosity
 */
static void
sscop_verbose(struct sscop *sscop __unused, void *arg __unused,
    const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fprintf(stderr, "\n");
}
static void
sscfu_verbose(struct sscfu *sscfu __unused, void *arg __unused,
    const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fprintf(stderr, "\n");
}

static void
user_lower_to_sscop(struct uni_msg *msg)
{

	sscop_input(stack.sscop, msg);
}

static void
user_manage_to_sscop(u_int sig, u_int arg1 __unused, u_int arg2 __unused,
    struct uni_msg *msg)
{
	sscop_maasig(stack.sscop, sig, msg);
}

static void
user_upper_to_sscfu(u_int sig, struct uni_msg *msg)
{
	if (sscfu_saalsig(stack.sscf, sig, msg))
		errx(1, "error in %s", __func__);
}

static const struct sscop_funcs sscop_funcs = {
	sscop_send_manage,
	sscop_send_upper,
	sscop_send_lower,
	sscop_verbose,
	sscop_start_timer,
	sscop_stop_timer
};

static const struct sscfu_funcs sscfu_funcs = {
	sscfu_send_upper,
	sscfu_send_lower,
	sscfu_window,
	sscfu_verbose
};

/*
 * Instantiate a user space stack
 */
static void
user_init(evContext *ctx)
{

	stack.ctx = ctx;

	if ((stack.sscop = sscop_create(&stack, &sscop_funcs)) == NULL)
		err(1, "cannot create sscop instance");

	if ((stack.sscf = sscfu_create(&stack, &sscfu_funcs)) == NULL)
		err(1, "cannot create sscfu instance");
}


const struct sscfu_stack user_stack = {
	user_init,
	user_debug,
	user_upper_to_sscfu,
	user_manage_to_sscop,
	user_lower_to_sscop,
	user_reset,
	user_start
};
